# This will allow shorter imports.
from SPC.Rule import Rule
from SPC.ControlChart import ControlChart
from SPC.ControlChartConstants import ControlChartConstants
from SPC.XbarRControlChartConstants import XbarRControlChartConstants
from SPC.XbarSControlChartConstants import XbarSControlChartConstants
from SPC.MultiVariChart import MultiVariChart
from SPC.ProcessCapabilityChart import ProcessCapabilityChart
from SPC.Rule01 import Rule01
from SPC.Rule02 import Rule02
from SPC.Rule03 import Rule03
from SPC.Rule04 import Rule04
from SPC.Rule05 import Rule05
from SPC.Rule06 import Rule06
from SPC.Rule07 import Rule07
from SPC.Rule08 import Rule08
from SPC.XbarmRControlChart import XbarmRControlChart
from SPC.XbarRControlChart import XbarRControlChart
from SPC.XbarSControlChart import XbarSControlChart
from SPC.XmRControlChart import XmRControlChart
from SPC.ImRControlChart import ImRControlChart
from SPC.UControlChart import UControlChart
from SPC.CControlChart import CControlChart
from SPC.PControlChart import PControlChart
from SPC.NPControlChart import NPControlChart
